#!/usr/bin/env python3
"""
Updates various web services using an updater yaml file

Usage:
  update-products [FILE] [options]
  
Options:
  -h --help         Show this screen.
  --answer=ANSWER   Answers ANSWER to any question instead of waiting for input

Arguments:
  FILE              The YAML config file to use. [default: ./.updater.yml]
"""
import os
import sys
from time import sleep as wait
from time import time as timestamp
from yaspin import yaspin
import yaml
from termcolor import cprint, colored
from subprocess import run
from PyInquirer import prompt, Separator
from typing import List, Tuple, Dict, Optional, Callable
from docopt import docopt

CHECKMARK = colored("√", 'green')

def main():
  args = docopt(__doc__)
  print(args)
  default_answer = args.get('--answer', None)
  
  templates, products, options, before, after = load_yaml_file(args['FILE'] or './.updater.yml')
  
  for step_name, command in iterate_steps(before):
    do_step(step_name, command, show=options['show stdout']['before'])

  for product in iterate_products(templates, products):
    update_project(default_answer, **product)

  for step_name, command in iterate_steps(after):
    do_step(step_name, command, show=options['show stdout']['after'])

def load_yaml_file(filepath):
  if not os.path.isfile(filepath):
    print("FATAL: Couldn't file a .updater.yml file in the current directory")
    sys.exit()
    
  with open(filepath, 'r') as file:
    raw = file.read()
    parsed = yaml.load(raw, Loader=yaml.SafeLoader)
    
  try:
    products = parsed['products']
    templates = parsed.get('templates', {})
    options = parsed.get('options', { 'show stdout': { 'before': False, 'after': True } })
    before = parsed.get('before', [])
    after = parsed.get('after', [])
  except KeyError:
    print('FATAL: Please set your products in .updater.yml')
    sys.exit()
  
  return templates, products, options, before, after
  

def iterate_products(
  templates: dict,
  products: List[dict]
):
  # The yaml is structured like:
  # products:
  # - Verbose product name:
  #     pm2: ...
  #     dir: ...
  #     ...
  cleaned_products = []
  for product in products:
    cleaned_product = {}
    for verbose_name, product in product.items():
      cleaned_product['verbose_name'] = verbose_name
      for key, val in product.items():
        val = _apply_template(key, val, templates)
        key = key.replace(' ', '_')
        cleaned_product[key] = val
      cleaned_products.append(cleaned_product)

  for product in cleaned_products:
    yield product

def _apply_template(
  key: str,
  value: str,
  templates: dict
):
  if key in templates.keys():
    return templates[key] % value
  else:
    return value

def update_project(
  default_answer: Optional[str],
  dir: str,
  pm2: Optional[str],
  clone_url: str,
  verbose_name: Optional[str], 
  steps: List[Tuple[str, str, Optional[Dict]]],
  **unused_keys
):
  def ask(*promptargs):
    if default_answer:
      ret = dict()
      ret[promptargs[0][0]['name']] = default_answer
      return ret
    else:
      return prompt(*promptargs)
  
  ## Print a header
  cprint('\n\n' + (verbose_name or pm2 or dir).upper() + '\n\n', 'yellow', attrs=['bold'])
  ## Create the dir & clone if it doesnt exist
  if not os.path.isdir(dir):
    if ask([{
      'type': 'confirm',
      'message': "The repository does not exist. Do I clone it?",
      'default': True,
      'name': 'clone'
    }])['clone']:
      git_clone(clone_url, dir)
    else:
      return
  with working_dir(dir):
    ## Check if the repo is up to date
    uptodate = is_up_to_date()
    ## Get choices (if the app is not managed by pm2, we can't show the "Just restart the app" option)
    what_to_do_choices = [
      'Nothing',
      'Do as if it was not up to date',
      Separator(),
    ]
    if pm2: what_to_do_choices.append('Just restart the app')
    for step_name, _ in iterate_steps(steps):
      what_to_do_choices.append(step_name)
    what_to_do = ''
    ## If the repo is up to date, ask what to do
    if uptodate:
      what_to_do = ask([{
        'type': 'list',
        'message': "The repository is already up to date. What should I do?",
        'name': 'what_to_do',
        'default': 'Nothing',
        'choices': what_to_do_choices
      }])['what_to_do']
      ## Exit
      if what_to_do == 'Nothing':
        return

    ## Pull from origin
    if not uptodate:
      git_pull()
      print(CHECKMARK + ' Pulled.')
    ## Do the steps
    if uptodate and what_to_do in [s[0] for s in iterate_steps(steps)]:
      for step_name, command in iterate_steps(steps):
        do_step(step_name, command)
    elif not uptodate or what_to_do == 'Do as if it was not up to date':
      update(steps)
    ## Restart (if the app is managed by pm2)
    if pm2:
      restart(pm2)
      print(CHECKMARK + ' Restarted.')


class working_dir:
  def __init__(self, path):
    self.path = path
    self.cwd = os.getcwd()
  def __enter__(self):
    os.chdir(self.path)
    cprint(f"Now at {os.getcwd()}", attrs=['dark'])
  def __exit__(self, type, value, traceback):
    os.chdir(self.cwd)
    cprint(f"Now at {os.getcwd()}", attrs=['dark'])

@yaspin(text="Checking for updates...", color="cyan")
def is_up_to_date():
  shell('sudo git fetch')
  stdout = shell('git --no-pager log HEAD..origin/master --oneline').stdout
  uptodate = stdout == b''
  return uptodate

@yaspin(text="Pulling from origin...", color="cyan")
def git_pull():
  shell('git pull')
  # print("Done.")

@yaspin(text="Cloning repository...", color="cyan")
def git_clone(git_clone_url, dir_name):
  shell(f'git clone {git_clone_url} {dir_name}')

@yaspin(text="Restarting...", color="cyan")
def restart(pm2_app):
  shell(f'pm2 restart {pm2_app}')

def update(steps):
  for spinner_text, command in iterate_steps(steps):
    do_step(spinner_text, command)

def do_step(text, command, **opts):
  @yaspin(text=text+'...', color='cyan')
  def doit():
    shell(command, **opts)
  doit()
  print(CHECKMARK + ' ' + _word_being_to_preterit(text, 'Done') + '.')

def iterate_steps(steps):
  for step in steps:
    for (verbose_name, command) in step.items():
      yield verbose_name, command

def _word_being_to_preterit(text: str, fallback: str):
  """
  Replaces the first word of :text: with its preterit tense, 
  or return :fallback:
  """
  REPLACE_MAP = {
    'Installing': 'Installed',
    'Building': 'Built',
    'Activating': 'Activated',
    'Copying': 'Copied',
    'Moving': 'Moved',
    'Compiling': 'Compiled',
    'Linting': 'Linted',
    'Stopping': 'Stopped',
    'Getting': 'Got',
    'Setting': 'Set',
    'Extracting': 'Extracted',
    'Restarting': 'Restarted',
    'Reloading': 'Reloaded',
    'Launching': 'Launched',
    'Applying': 'Applied'
  }
  for (being, preterit) in REPLACE_MAP.items():
    if being in text.split(' '):
      return text.replace(being, preterit)
  return fallback

def shell(command, show=False):
  return run(command, shell=True, capture_output=not show)

if __name__ == "__main__":
  main()
